/***************************************************************************
 *   Copyright (C) 2004 by David Banz                                      *
 *   neko@netcologne.de                                                    *
 *   GPL'ed                                                                *
 ***************************************************************************/

#include <stdint.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include "player.h"
void tfmxIrqIn();

char act[8]={1,1,1,1,1,1,1,1};
extern struct Hdb hdb[8];
extern struct Cdb cdb[8];
extern struct Mdb mdb;
#define HALFBUFSIZE 65536
#define BUFSIZE 131072 

#define bofsize 4096
uint32_t bqueue = 0;

unsigned long bytes=0;

U32 blocksize = (HALFBUFSIZE/8);
U32 stereo=0;

int eRem=0; /* remainder of eclocks */
int blend=1; /* default to blended mode */
int filt=1; /* light lpf */
void tfmxIrqIn(void);

/* Simple little three-position weighted-sum LPF. */
void filter(S32 *b, int num)
{
	register int x;
	static int wl=0,wr=0; /* actually backwards but who cares? */
	switch(filt&3)
	{
	case 3:
		for (x=0;x<num;x++)
		{
			wl=((b[HALFBUFSIZE])+wl*3)>>2; b[HALFBUFSIZE]=wl;
			wr=((*b)+wr*3)>>2; *b++=wr;
		}
		break;
	case 2:
		for (x=0;x<num;x++)
		{
			wl=((b[HALFBUFSIZE])+wl)>>1; b[HALFBUFSIZE]=wl;
			wr=((*b)+wr)>>1; *b++=wr;
		}
		break;
	case 1:
		for (x=0;x<num;x++)
		{
			wl=((b[HALFBUFSIZE])*3+wl)>>2; b[HALFBUFSIZE]=wl;
			wr=((*b)*3+wr)>>2; *b++=wr;
		}
		break;
	case 0:
		break;
	}
}

/* This one looks like a good candidate for high optimization... */
void stereoblend(S32 *b, uint32_t num)
{
    uint32_t x;
	for (x=0;x<num;x++) {
	    register int y0,y1;
		y0=((b[x+HALFBUFSIZE]*11)+(b[x]* 5))>>4;
        y1=((b[x+HALFBUFSIZE]* 5)+(b[x]*11))>>4;
        b[x]=y1;
        b[x+HALFBUFSIZE]=y0;
	}
}

void conv_s16(S16 *buf, S32 *b, U32 num)
{
	S16 *a=(S16 *)buf;
	S32 *c=b;
	int x;

    if (blend)
	    stereoblend(b,num);

	if (stereo) {
		for (x=0;x<num;x++) {
            a[2*x  ]=(b[HALFBUFSIZE]);
            a[2*x+1]=(b[0]);
            b++;
		}
	} else {
		for (x=0;x<num;x++) {
/* reverted, the new version broke -b0 */
			a[x]=((b[HALFBUFSIZE]+b[0])>>1);
            b++;
/*			*a++ = b[HALFBUFSIZE];
			*a = *a + *b++;
			*a = *a >> 1;*/
		}
	}

	for(x=0;x<num;x++) {
		c[HALFBUFSIZE]=0;
		*c++=0;
	}
}

static int nul=0;
void (*mix)(struct Hdb *,int,S32 *);

void mix_add(struct Hdb *hw,int n,S32 *b)
{
	register S8 * p = hw->sbeg;
	register U32 ps=hw->pos;
	int v=hw->vol;
	U32 d=hw->delta;
	U32 l=(hw->slen<<14);

	if (v>0x40)v=0x40;

/* This used to have (p==&smplbuf).  Broke with GrandMonsterSlam */
	if ((p==(S8 *)&nul)||( ((hw->mode)&1)==0 )||(l<0x10000))
		return;

	if ((hw->mode&3)==1) {
		p= hw->sbeg=hw->SampleStart;
		l=(hw->slen=hw->SampleLength)<<14;
		ps=0;
		hw->mode|=2;
/*		hw->loop(&hw);*/
	}

	if (!v)
	{
#if 0		/* Will be supported someday... */
		while(n--){
			(*b++)+=(p[(ps+=d)>>14]*v);
			if (ps<l) continue;
			ps-=l;
			p=hw->SampleStart;
			if (((l=hw->SampleLength<<14)<=0x10000) ||
			    (!hw->loop(hw)) )
					{
				ps=l=d=0;
				p=smplbuf;
				break;
			}
		}
		return;
#endif
	}

	while(n--){
		int smpl = p[(ps+=d)>>14];
		(*b++)+=(v*smpl);
		if (ps<l) continue;
		ps-=l;
		p=hw->SampleStart;
		if (((l=((hw->slen=hw->SampleLength)<<14))<0x10000) ||
		     (!hw->loop(hw))) {
			hw->slen=ps=d=0;
			p=smplbuf;
			break;
		}
	}

	hw->sbeg=p;
	hw->pos=ps;
	hw->delta=d;
	if (hw->mode&4) (hw->mode=0);
}

void mix_add_ov(struct Hdb *hw,int n,S32 *b)
{
	register S8 * p = hw->sbeg;
	register U32 ps=hw->pos;
	register U32 psreal;
	int v=hw->vol;
	U32 d=hw->delta;
	U32 l=(hw->slen<<14);
    int smpl,v1,v2;

	if (v>0x40)v=0x40;

/* This used to have (p==&smplbuf).  Broke with GrandMonsterSlam */
	if ((p==(S8 *)&nul)||( ((hw->mode)&1)==0 )||(l<0x10000))
		return;

	if ((hw->mode&3)==1) {
		p= hw->sbeg=hw->SampleStart;
		l=(hw->slen=hw->SampleLength)<<14;
		ps=0;
		hw->mode|=2;
	/*	hw->loop(&hw); */
	}

	if (!v)
	{
#if 0		/* Will be supported someday... */
		while(n--){
			(*b++)+=(p[(ps+=d)>>14]*v);
			if (ps<l) continue;
			ps-=l;
			p=hw->SampleStart;
			if (((l=hw->SampleLength<<14)<=0x10000) ||
			    (!hw->loop(hw)) )
					{
				ps=l=d=0;
				p=smplbuf;
				break;
			}
		}
		return;
#endif
	}

/*
#   define RESAMPLATION \
      v1=src[ofs>>FRACTION_BITS];\
      v2=src[(ofs>>FRACTION_BITS)+1];\
      *dest++ = v1 + (((v2-v1) * (ofs & FRACTION_MASK)) >> FRACTION_BITS);

*/
#define FRACTION_BITS 14
#define INTEGER_MASK (0xFFFFFFFF << FRACTION_BITS)
#define FRACTION_MASK (~ INTEGER_MASK)

	while(n--){
		psreal = ps>>FRACTION_BITS;
		v1 = p[psreal];
		if (psreal+1 < hw->slen) {
			v2 = p[psreal+1];
		} else {
			v2 = hw->SampleStart[0];
		}
		smpl = ((v1 + (((int) ((v2-v1) * (ps & FRACTION_MASK))) >> FRACTION_BITS)));
		(*b++) += v*smpl;
		ps += d;

		if (ps<l) continue;
		ps-=l;
		p=hw->SampleStart;
		if (((l=((hw->slen=hw->SampleLength)<<14))<0x10000) ||
		     (!hw->loop(hw))) {
			hw->slen=ps=d=0;
			p=smplbuf;
			break;
		}
	}

	hw->sbeg=p;
	hw->pos=ps;
	hw->delta=d;
	if (hw->mode&4) (hw->mode=0);
}

void (*mix)(struct Hdb *,int,S32 *)=&mix_add_ov;

void mixit(S32 *tbuf, U32 n, U32 b)
{
	int x;
	S32 *y;
	//if (multimode) {
		if(act[4])mix(&hdb[4],n,&tbuf[HALFBUFSIZE+b]);
		if(act[5])mix(&hdb[5],n,&tbuf[b]);
		if(act[6])mix(&hdb[6],n,&tbuf[b]);
		if(act[7])mix(&hdb[7],n,&tbuf[b]);
		y=&tbuf[HALFBUFSIZE+b];
		for (x=0;x<n;x++,y++) {
            int smpl=*y;
			*y=(smpl>16383)?16383:(smpl<-16383)?-16383:smpl;
        }
	//} else {
		//if(act[3])mix(&hdb[3],n,&tbuf[b]);
    //}
	if(act[3])mix(&hdb[3],n,&tbuf[b]);
	if(act[0])mix(&hdb[0],n,&tbuf[b]);
	if(act[1])mix(&hdb[1],n,&tbuf[HALFBUFSIZE+b]);
	if(act[2])mix(&hdb[2],n,&tbuf[HALFBUFSIZE+b]);
}

int try_to_makeblock(S16 *buf)
{
	static S32 nb=0,bd=0; /* num bytes, bytes done */
    S32 tbuf[HALFBUFSIZE*2];
	unsigned int multiplier=(stereo?1:2);
	int n,r=0;

	while ((unsigned int)(!r) && 
		   ((unsigned int)(bqueue+2) < (8*multiplier)))
	{
		tfmxIrqIn();
		nb=(eClocks*(outRate>>1));
		eRem+=(nb%357955);
		nb/=357955;
		if (eRem>357955) nb++,eRem-=357955;
		while (nb>0) {
			n=blocksize-bd;
			if (n>nb) n=nb;
			mixit(tbuf, n, bd);
			bytes+=n;
			bd+=n;
			nb-=n;
			if (((unsigned int)bd) == blocksize) {
	            //filter(&tbuf[0], bd);
				//conv_s16(buf, &tbuf[0], bd);
	            //filter(tbuf, bd);
				conv_s16(buf, tbuf, bd);
				bd=0;
				bqueue++;
				/*printf("make %d\n",bqueue);*/ r++;
			}
		}
	}
	return((mdb.PlayerEnable)?r:-1);
}

